# pymoode imports
from pymoode.survival.replacement import ImprovementReplacement
from pymoode.algorithms.base.differential import DifferentialEvolution


# =========================================================================================================
# Implementation
# =========================================================================================================

class DE(DifferentialEvolution):

    def __init__(self,
                 pop_size=100,
                 variant="DE/rand/1/bin",
                 CR=0.7,
                 F=(0.5, 1.0),
                 gamma=1e-4,
                 de_repair="bounce-back",
                 survival=ImprovementReplacement(),
                 **kwargs):
        """
        Single-objective Differential Evolution proposed by Storn and Price (1997).

        Storn, R. & Price, K., 1997. Differential evolution-a simple and efficient heuristic for global optimization over continuous spaces. J. Glob. Optim., 11(4), pp. 341-359.

        Parameters
        ----------
        pop_size : int, optional
            Population size. Defaults to 100.

        variant : str, optional
            Differential evolution strategy. Must be a string in the format: "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'. Selection variants are:

                - 'ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'

            The selection strategy 'ranked' might be helpful to improve convergence speed without much harm to diversity. Defaults to 'DE/rand/1/bin'.

        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.

        genetic_mutation, optional
            Pymoo's genetic mutation operator after crossover. Defaults to NoMutation().

        survival : Survival, optional
            Replacement survival operator. Defaults to ImprovementReplacement().

        repair : Repair, optional
            Pymoo's repair operator after mutation. Defaults to NoRepair().
        """
        
        super().__init__(
            pop_size=pop_size,
            variant=variant,
            CR=CR,
            F=F,
            gamma=gamma,
            de_repair=de_repair,
            survival=survival,
            **kwargs,
        )




# Native
from typing import Optional

# External
import numpy as np

# pymoo imports
from pymoo.operators.sampling.lhs import LHS
from pymoo.termination.default import DefaultSingleObjectiveTermination, DefaultMultiObjectiveTermination
from pymoo.util.display.single import SingleObjectiveOutput
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.misc import has_feasible

# pymoode imports
from pymoode.operators.variant import DifferentialVariant
from pymoode.algorithms.base.evolutionary import EvolutionaryAlgorithm
from pymoode.survival.replacement import ImprovementReplacement
from pymoode.survival import RankAndCrowding


# =========================================================================================================
# Implementation
# =========================================================================================================

class DifferentialEvolution(EvolutionaryAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=LHS(),
                 variant="DE/rand/1/bin",
                 CR=0.7,
                 F=(0.5, 1.0),
                 gamma=1e-4,
                 de_repair="bounce-back",
                 survival=ImprovementReplacement(),
                 advance_after_initial_infill=True,
                 output=SingleObjectiveOutput(),
                 **kwargs):
        """
        Base class for Differential Evolution algorithms

        Parameters
        ----------
        pop_size : int, optional
            Population size. Defaults to 100.

        sampling : Sampling, optional
            Sampling strategy of pymoo. Defaults to LHS().

        variant : str, optional
            Differential evolution strategy. Must be a string in the format: "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'. Selection variants are:

                - 'ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'

            The selection strategy 'ranked' might be helpful to improve convergence speed without much harm to diversity. Defaults to 'DE/rand/1/bin'.

        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.

        genetic_mutation, optional
            Pymoo's genetic mutation operator after crossover. Defaults to NoMutation().

        survival : Survival, optional
            Replacement survival operator. Defaults to ImprovementReplacement().

        repair : Repair, optional
            Pymoo's repair operator after mutation. Defaults to NoRepair().
        """

        # Mating
        mating = DifferentialVariant(
            variant=variant, CR=CR, F=F, gamma=gamma,
            de_repair=de_repair, **kwargs,
        )

        # Number of offsprings at each generation
        n_offsprings = pop_size

        super().__init__(
            pop_size=pop_size,
            sampling=sampling,
            mating=mating,
            n_offsprings=n_offsprings,
            eliminate_duplicates=None,
            survival=survival,
            output=output,
            advance_after_initial_infill=advance_after_initial_infill,
            **kwargs,
        )

        self.termination = DefaultSingleObjectiveTermination()

    def _initialize_advance(self, infills=None, **kwargs):
        self.pop = self.survival.do(self.problem, infills, None, n_survive=self.pop_size)

    def _infill(self):
        infills = self.mating(self.problem, self.pop, self.n_offsprings, algorithm=self)
        return infills

    def _advance(self, infills=None, **kwargs):

        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must be provided."

        # One-to-one replacement survival
        self.pop = self.survival.do(self.problem, self.pop, infills)

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.pop[self.pop.get("rank") == 0]


class MODE(DifferentialEvolution):

    def __init__(self,
                 pop_size=100,
                 sampling=LHS(),
                 variant="DE/rand/1/bin",
                 CR=0.5,
                 F=None,
                 gamma=0.0001,
                 de_repair="bounce-back",
                 survival=RankAndCrowding(),
                 output=MultiObjectiveOutput(),
                 **kwargs):

        super().__init__(
            pop_size=pop_size,
            sampling=sampling,
            variant=variant,
            CR=CR,
            F=F,
            gamma=gamma,
            de_repair=de_repair,
            survival=survival,
            output=output,
            **kwargs,
        )

        self.termination = DefaultMultiObjectiveTermination()

    def _initialize_advance(self, infills=None, **kwargs):
        self.pop = self.survival.do(self.problem, infills, None, n_survive=self.pop_size)



from pymoode.algorithms.de import DifferentialEvolution
from pymoode.algorithms.de import DE
from pymoode.algorithms.gde3 import GDE3, GDE3MNN, GDE32NN, GDE3P
from pymoode.algorithms.nsde import NSDE
from pymoode.algorithms.nsder import NSDER



# Native
from abc import abstractmethod
from typing import Optional, Union

# External
import numpy as np

# pymoo
from pymoo.core.crossover import Crossover
from pymoo.core.population import Population
from pymoo.core.problem import Problem


# =========================================================================================================
# Implementation
# =========================================================================================================

class DifferentialOperator(Crossover):

    def __init__(self, n_parents=None, **kwargs):
        """White label for differential evolution operators

        Parameters
        ----------
        n_parents : int | None, optional
            Number of parents necessary in its operations. Useful for compatibility with pymoo.
        """
        # __init__ operator
        super().__init__(n_parents=n_parents, n_offsprings=1, prob=1.0, **kwargs)

    @staticmethod
    def default_prepare(pop: Population, parents: Union[Population, np.ndarray]):
        """Utility function that converts population and parents from pymoo Selection to pop and X

        Parameters
        ----------
        pop : Population
            pymoo population

        parents : Population | np.ndarray (n_samples, n_parents) | None
            Parent population or indices

        Returns
        -------
        pop, X : Population (n_samples, n_parents), np.ndarray (n_parents, n_samples, n_var)
            Population and corresponding decision variables
        """
        # Convert pop if parents is not None
        if parents is not None:
            pop = pop[parents]

        # Get all X values for mutation parents
        X = np.swapaxes(pop, 0, 1).get("X")
        return pop, X

    @abstractmethod
    def do(
        self,
        problem: Problem,
        pop: Population,
        parents: Optional[Union[Population, np.ndarray]] = None,
        **kwargs
    ):
        pass

    @abstractmethod
    def _do(self, problem: Problem, X: np.ndarray, **kwargs):
        pass



import numpy as np


def calc_pcd(X: np.ndarray, n_remove=0):

    N = X.shape[0]
    M = X.shape[1]

    if n_remove <= (N - M):
        if n_remove < 0:
            n_remove = 0
        else:
            pass
    else:
        n_remove = N - M

    extremes_min = np.argmin(X, axis=0)
    extremes_max = np.argmax(X, axis=0)

    min_vals = np.min(X, axis=0)
    max_vals = np.max(X, axis=0)

    extremes = np.concatenate((extremes_min, extremes_max))

    X = (X - min_vals) / (max_vals - min_vals)

    H = np.arange(N)
    d = np.full(N, np.inf)

    I = np.argsort(X, axis=0, kind='mergesort')

    # sort the objective space values for the whole matrix
    _X = X[I, np.arange(M)]

    # calculate the distance from each point to the last and next
    dist = np.row_stack([_X, np.full(M, np.inf)]) - np.row_stack([np.full(M, -np.inf), _X])

    # prepare the distance to last and next vectors
    dist_to_last, dist_to_next = dist, np.copy(dist)
    dist_to_last, dist_to_next = dist_to_last[:-1], dist_to_next[1:]

    # if we divide by zero because all values in one columns are equal replace by none
    dist_to_last[np.isnan(dist_to_last)] = 0.0
    dist_to_next[np.isnan(dist_to_next)] = 0.0

    # sum up the distance to next and last and norm by objectives - also reorder from sorted list
    J = np.argsort(I, axis=0)
    _d = np.sum(dist_to_last[J, np.arange(M)] + dist_to_next[J, np.arange(M)], axis=1)
    d[H] = _d
    d[extremes] = np.inf

    n_removed = 0

    #While n_remove not acheived
    while n_removed < (n_remove - 1):

        #Obtain element to drop
        _d = d[H]
        _k = np.argmin(_d)
        k = H[_k]

        H = H[H != k]

        #Update index
        n_removed = n_removed + 1

        I = np.argsort(X[H].copy(), axis=0, kind='mergesort')

        # sort the objective space values for the whole matrix
        _X = X[H].copy()[I, np.arange(M)]

        # calculate the distance from each point to the last and next
        dist = np.row_stack([_X, np.full(M, np.inf)]) - np.row_stack([np.full(M, -np.inf), _X])

        # prepare the distance to last and next vectors
        dist_to_last, dist_to_next = dist, np.copy(dist)
        dist_to_last, dist_to_next = dist_to_last[:-1], dist_to_next[1:]

        # if we divide by zero because all values in one columns are equal replace by none
        dist_to_last[np.isnan(dist_to_last)] = 0.0
        dist_to_next[np.isnan(dist_to_next)] = 0.0

        # sum up the distance to next and last and norm by objectives - also reorder from sorted list
        J = np.argsort(I, axis=0)
        _d = np.sum(dist_to_last[J, np.arange(M)] + dist_to_next[J, np.arange(M)], axis=1)
        d[H] = _d
        d[extremes] = np.inf

    # Divide by number of objectives to obtain correct metric
    d = d / M

    return d




# pymoo imports
from pymoode.algorithms.base.evolutionary import EvolutionaryAlgorithm
from pymoo.core.mating import Mating
from pymoo.core.duplicate import DefaultDuplicateElimination, NoDuplicateElimination


# =========================================================================================================
# Implementation
# =========================================================================================================

class GeneticAlgorithm(EvolutionaryAlgorithm):
    
    def __init__(self,
                 pop_size=None,
                 sampling=None,
                 selection=None,
                 crossover=None,
                 mutation=None,
                 survival=None,
                 n_offsprings=None,
                 eliminate_duplicates=True,
                 repair=None,
                 **kwargs):
        """Base class for Genetic Algorithms. A Mating operator is instantiated using 
        selection, crossover, mutation, repair, and eliminate_duplicates arguments.

        Parameters
        ----------
        pop_size : int, optional
            Population size, by default None
        
        sampling : Sampling, optional
            pymoo Sampling instance, by default None
        
        selection : Selection, optional
            pymoo parent selection operator, by default None
        
        crossover : Crossover, optional
            pymoo crossover operator, by default None
        
        genetic_mutation, optional
            pymoo mutation operator, by default None
        
        survival : Survival, optional
            pymoo survival operator, by default None
        
        n_offsprings : int, optional
            Number of offspring individuals created at each generation, by default None
        
        eliminate_duplicates : DuplicateElimination | bool | None, optional
            Eliminate duplicates in mating, by default True
        
        repair : Repair, optional
            pymoo repair operator which should be passed to Mating and self. By default None
        
        advance_after_initial_infill : bool, optional
            Either or not apply survival after initialization, by default False
        """
        
        # set the duplicate detection class - a boolean value chooses the default duplicate detection
        if isinstance(eliminate_duplicates, bool):
            if eliminate_duplicates:
                eliminate_duplicates = DefaultDuplicateElimination()
            else:
                eliminate_duplicates = NoDuplicateElimination()
        
        # Mating operator of genetic algorithms
        mating = Mating(selection, crossover, mutation,
                        repair=repair,
                        eliminate_duplicates=eliminate_duplicates)
        
        # Instantiate generic evolutionary algorithm
        super().__init__(
            pop_size=pop_size,
            sampling=sampling,
            survival=survival,
            mating=mating,
            n_offsprings=n_offsprings,
            eliminate_duplicates=eliminate_duplicates,
            repair=repair, 
            **kwargs,
        )



# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True

import numpy as np

from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.set cimport set as cpp_set


cdef extern from "math.h":
    double HUGE_VAL


# Returns elements to remove based on crowding metric d and heap of remaining elements H
cdef inline int c_get_drop(double[:] d, cpp_set[int] H):

    cdef:
        int i, min_i
        double min_d

    min_d = HUGE_VAL
    min_i = 0

    for i in H:

        if d[i] <= min_d:
            min_d = d[i]
            min_i = i

    return min_i


# Returns vector of positions of minimum values along axis 0 of a 2d memoryview
cdef inline vector[int] c_get_argmin(double[:, :] X):

    cdef:
        int N, M, min_i, n, m
        double min_val
        vector[int] indexes

    N = X.shape[0]
    M = X.shape[1]

    indexes = vector[int]()

    for m in range(M):

        min_i = 0
        min_val = X[0, m]

        for n in range(N):

            if X[n, m] < min_val:

                min_i = n
                min_val = X[n, m]

        indexes.push_back(min_i)

    return indexes


# Returns vector of positions of maximum values along axis 0 of a 2d memoryview
cdef inline vector[int] c_get_argmax(double[:, :] X):

    cdef:
        int N, M, max_i, n, m
        double max_val
        vector[int] indexes

    N = X.shape[0]
    M = X.shape[1]

    indexes = vector[int]()

    for m in range(M):

        max_i = 0
        max_val = X[0, m]

        for n in range(N):

            if X[n, m] > max_val:

                max_i = n
                max_val = X[n, m]

        indexes.push_back(max_i)

    return indexes


# Performs normalization of a 2d memoryview
cdef inline double[:, :] c_normalize_array(double[:, :] X, vector[int] extremes_max, vector[int] extremes_min):

    cdef:
        int N = X.shape[0]
        int M = X.shape[1]
        int n, m, l, u
        double l_val, u_val, diff_val
        vector[double] min_vals, max_vals

    min_vals = vector[double]()
    max_vals = vector[double]()

    m = 0
    for u in extremes_max:
        u_val = X[u, m]
        max_vals.push_back(u_val)
        m = m + 1

    m = 0
    for l in extremes_min:
        l_val = X[l, m]
        min_vals.push_back(l_val)
        m = m + 1

    for m in range(M):

        diff_val = max_vals[m] - min_vals[m]
        if diff_val == 0.0:
            diff_val = 1.0

        for n in range(N):

            X[n, m] = (X[n, m] - min_vals[m]) / diff_val

    return X



# Native
from typing import Optional

# External
import numpy as np

# pymoo imports
from pymoo.core.duplicate import DefaultDuplicateElimination
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.problem import Problem
from pymoo.core.survival import Survival
from pymoo.core.individual import Individual

# pymoode imports
from pymoode.survival.fitness import FitnessSurvival


# =========================================================================================================
# Implementation
# =========================================================================================================

def is_better(_new: Individual, _old: Individual, eps=0.0):
    both_infeasible = not _old.feasible[0] and not _new.feasible[0]
    both_feasible = _old.feasible[0] and _new.feasible[0]

    if both_infeasible and _old.CV[0] - _new.CV[0] > eps:
        return True
    elif not _old.feasible and _new.feasible:
        return True
    elif both_feasible and _old.F[0] - _new.F[0] > eps:
        return True

    return False


class BaseReplacement(Survival):

    def __init__(
        self,
        fitness: Optional[Survival] = None
    ):
        """Base class for Replacement survival

        Parameters
        ----------
        fitness : Survival, optional
            Survival used in ranking and fitness assignement, by default None
        """
        super().__init__(filter_infeasible=False)
        if fitness is None:
            fitness = FitnessSurvival()
        self.fitness = fitness

    def do(
        self,
        problem: Problem,
        pop: Population,
        off: Population,
        return_indices=False,
        inplace=False,
        **kwargs
    ):
        """Given a problem a parent population and offspring canditates return next generation.
        By default, only fitness assigment is performed on parents. To define which ones should be replaced,
        define method ``_do`` in child classes that returns a boolean if parent should be replaced.

        Parameters
        ----------
        problem : Problem
            Pymoo problem

        pop : Population
            Parent population

        off : Population | None
            Offspring candidates. If None, only fitness assigment is performed on parents.

        return_indices : bool, optional
            Either or not to just return boolean of positions to be replaced, by default False

        inplace : bool, optional
            Change population inplace, by default False

        Returns
        -------
        Population
            Population that proceeds into the next generation
        """

        # If no offspring is available just do fitness assignment
        if (off is None) or len(off) == 0:
            return self.fitness.do(problem, pop, n_survive=None)

        # this makes it usable as a traditional survival
        elif isinstance(off, int) or isinstance(off, np.integer):
            k = off
            off = pop[k:]
            pop = pop[:k]

        assert len(pop) == len(off), "For the replacement pop and off must have the same number of individuals."

        pop = Population.create(pop) if isinstance(pop, Individual) else pop
        off = Population.create(off) if isinstance(off, Individual) else off

        I = self._do(problem, pop, off, **kwargs)

        if return_indices:
            return I
        else:
            if not inplace:
                pop = pop.copy()
            pop[I] = off[I]
            return self.fitness.do(problem, pop, n_survive=None)

    def _do(
        self,
        problem: Problem,
        pop: Population,
        off: Population,
        **kwargs
    ):
        return np.zeros(len(pop), dtype=np.bool_)


class ImprovementReplacement(BaseReplacement):

    def do(
        self,
        problem: Problem,
        pop: Population,
        off: Population,
        return_indices=False,
        inplace=False,
        **kwargs
    ):
        """Given a problem a parent population and offspring canditates return next generation,
        it selects those that proceed into the next generation via one-to-one comparison.
        Feasible solutions are always preferred to infeasible and infeasible solutions are compared by
        overall constraint violation.

        Parameters
        ----------
        problem : Problem
            Pymoo problem

        pop : Population
            Parent population

        off : Population | None
            Offspring candidates. If None, only fitness assigment is performed on parents.

        return_indices : bool, optional
            Either or not to just return boolean of positions to be replaced, by default False

        inplace : bool, optional
            Change population inplace, by default False

        Returns
        -------
        Population
            Population that proceeds into the next generation
        """
        return super().do(problem, pop, off, return_indices=return_indices, inplace=inplace, **kwargs)

    def _do(
        self,
        problem: Problem,
        pop: Population,
        off: Population,
        **kwargs
    ):

        ret = np.full((len(pop), 1), False)

        pop_F, pop_CV, pop_feas = pop.get("F", "CV", "feasible")
        off_F, off_CV, off_feas = off.get("F", "CV", "feasible")

        if problem.has_constraints() > 0:

            # 1) Both infeasible and constraints have been improved
            ret[(~pop_feas & ~off_feas) & (off_CV < pop_CV)] = True

            # 2) A solution became feasible
            ret[~pop_feas & off_feas] = True

            # 3) Both feasible but objective space value has improved
            ret[(pop_feas & off_feas) & (off_F < pop_F)] = True

        else:
            ret[off_F < pop_F] = True

        # never allow duplicates to become part of the population when replacement is used
        _, _, is_duplicate = DefaultDuplicateElimination(epsilon=0.0).do(off, pop, return_indices=True)
        ret[is_duplicate] = False

        return ret[:, 0]


def parameter_less(f, cv):
    v = np.copy(f)
    infeas = cv > 0
    v[infeas] = f.max() + cv[infeas]
    return v


def hierarchical_sort(f, cv=None):
    if cv is not None:
        f = parameter_less(f, cv)
    return np.argsort(f)



import argparse
from distutils.errors import CompileError
import os
import sys
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext


# ---------------------------------------------------------------------------------------------------------
# BASE CONFIGURATION
# ---------------------------------------------------------------------------------------------------------

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

BASE_PACKAGE = 'pymoode'

base_kwargs = dict(
    name = 'pymoode',
    packages = [BASE_PACKAGE] + [f"{BASE_PACKAGE}." + e for e in find_packages(where=BASE_PACKAGE)],
    version = '0.3.0',
    license='Apache License 2.0',
    description = 'A Python optimization package using Differential Evolution.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author = 'Bruno Scalia C. F. Leite',
    author_email = 'mooscaliaproject@gmail.com',
    url = 'https://github.com/mooscaliaproject/pymoode',
    download_url = 'https://github.com/mooscaliaproject/pymoode',
    keywords = [
            'Differential Evolution',
            'Multi-objective optimization',
            'GDE3',
            'NSDE',
            'NSDE-R',
            'NSGA-II',
            'Genetic Algorithm',
            'Crowding Distances',
            'Evolutionary Algorithms',
            'Evolutionary Optimization',
            'Metaheuristics',
        ],
    install_requires=[
            'numpy>=1.19.0',
            'pymoo==0.6.1',
            'scipy>=1.7.0',
            'future',
        ],
)

# ---------------------------------------------------------------------------------------------------------
# ARGPARSER
# ---------------------------------------------------------------------------------------------------------

parser = argparse.ArgumentParser()

parser.add_argument('--nopyx', dest='nopyx', action='store_true', help='Whether the pyx files shall be considered at all.')
parser.add_argument('--nocython', dest='nocython', action='store_true', help='Whether pyx files shall be cythonized.')
parser.add_argument('--nolibs', dest='nolibs', action='store_true', help='Whether the libraries should be compiled.')
parser.add_argument('--markcython', dest='markcython', action='store_true', help='Whether to mark the html cython files.')

params, _ = parser.parse_known_args()
sys.argv = [arg for arg in sys.argv if not arg.lstrip("-") in params]

# ---------------------------------------------------------------------------------------------------------
# ADD MARKED HTML FILES FOR CYTHON
# ---------------------------------------------------------------------------------------------------------

if params.markcython:
    import Cython.Compiler.Options
    Cython.Compiler.Options.annotate = True

# ---------------------------------------------------------------------------------------------------------
# CLASS TO BUILD EXTENSIONS
# ---------------------------------------------------------------------------------------------------------

# exception that is thrown when the build fails
class CompilingFailed(Exception):
    pass


# try to compile, if not possible throw exception
def construct_build_ext(build_ext):
    class WrappedBuildExt(build_ext):
        def run(self):
            try:
                build_ext.run(self)
            except BaseException as e:
                raise CompilingFailed(e)

        def build_extension(self, ext):
            try:
                build_ext.build_extension(self, ext)
            except BaseException as e:
                raise CompilingFailed(e)

    return WrappedBuildExt

# ---------------------------------------------------------------------------------------------------------
# HANDLING CYTHON FILES
# ---------------------------------------------------------------------------------------------------------

ROOT = os.path.dirname(os.path.realpath(__file__))

if params.nopyx:
    ext_modules = []

else:
    try:
        if params.nocython:
            path = os.path.join(ROOT, "pymoode", "cython")
            pyx = [os.path.join(path, f) for f in os.listdir() if f.endswith(".pyx")]
            ext_modules = [Extension(f"pymoode.cython.{source[:-4]}", [source]) for source in pyx]
        else:
            try:
                from Cython.Build import cythonize
                ext_modules = cythonize("pymoode/cython/*.pyx")
            except ImportError:
                print('*' * 75)
                print("No Cython package found to convert .pyx files.")
                print("If no compilation occurs, .py files will be used instead, which provide the same results but with worse computational time.")
                print('*' * 75)
                ext_modules = []
    except:
        print('*' * 75)
        print("Problems compiling .pyx files.")
        print("If no compilation occurs, .py files will be used instead, which provide the same results but with worse computational time.")
        print('*' * 75)
        ext_modules = []

if not params.nolibs:

    if len(ext_modules) > 0:
        base_kwargs['ext_modules'] = ext_modules

        try:
            import numpy as np
            base_kwargs['include_dirs'] = [np.get_include()]

        except BaseException:
            raise CompileError(
                "NumPy libraries must be installed for compiled extensions! Speedups are not enabled."
            )

        base_kwargs['cmdclass'] = dict(build_ext=construct_build_ext(build_ext))

    else:
        print('*' * 75)
        print("External cython modules found.")
        print("To verify compilation success run:")
        print("from pymoode.survival.rank_and_crowding.metrics import IS_COMPILED")
        print("This variable will be True to mark compilation success;")
        print("If no compilation occurs, .py files will be used instead, which provide the same results but with worse computational time.")
        print('*' * 75)

# ---------------------------------------------------------------------------------------------------------
# RUN SETUP
# ---------------------------------------------------------------------------------------------------------

compiled_kwargs = base_kwargs.copy()
compiled_kwargs["ext_modules"] = ext_modules

try:
    setup(**compiled_kwargs)
    print('*' * 75)
    print("Installation successful at the first attempt.")
    print("To verify compilation success run:")
    print("from pymoode.survival.rank_and_crowding.metrics import IS_COMPILED")
    print('*' * 75)
except:
    print('*' * 75)
    print("Running setup with cython compilation failed.")
    print("Attempt to a pure Python setup.")
    print("If no compilation occurs, .py files will be used instead, which provide the same results but with worse computational time.")
    print('*' * 75)
    setup(**base_kwargs)



# Native
import warnings

# External
import numpy as np

# pymoo imports
from pymoo.operators.mutation.nom import NoMutation
from pymoo.core.infill import InfillCriterion
from pymoo.core.population import Population
from pymoo.core.problem import Problem

# pymoode imports
from pymoode.operators.des import DES
from pymoode.operators.dex import DEX
from pymoode.operators.dem import DEM


# =========================================================================================================
# Implementation
# =========================================================================================================

class DifferentialVariant(InfillCriterion):

    def __init__(self,
                 variant="DE/rand/1/bin",
                 CR=0.7,
                 F=(0.5, 1.0),
                 gamma=1e-4,
                 de_repair="bounce-back",
                 genetic_mutation=None,
                 **kwargs):
        """InfillCriterion class for Differential Evolution

        Parameters
        ----------
        variant : str, optional
            Differential evolution strategy. Must be a string in the format: "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'. Selection variants are:

                - 'ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'

            The selection strategy 'ranked' might be helpful to improve convergence speed without much harm to diversity. Defaults to 'DE/rand/1/bin'.

        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.

        genetic_mutation : Mutation, optional
            Pymoo's genetic algorithm's mutation operator after crossover. Defaults to NoMutation().

        repair : Repair, optional
            Pymoo's repair operator after mutation. Defaults to NoRepair().
        """

        # Fix deprecated pm kwargs
        kwargs, genetic_mutation = _fix_deprecated_pm_kwargs(kwargs, genetic_mutation)

        # Default initialization of InfillCriterion
        super().__init__(eliminate_duplicates=None, **kwargs)

        # Parse the information from the string
        _, selection_variant, n_diff, crossover_variant, = variant.split("/")
        n_diffs = int(n_diff)

        # When "to" in variant there are more than 1 difference vectors
        if "-to-" in variant:
            n_diffs =  n_diffs + 1

        # Define parent selection operator
        self.selection = DES(selection_variant)

        # Define differential evolution mutation
        self.de_mutation = DEM(F=F, gamma=gamma, de_repair=de_repair, n_diffs=n_diffs)

        # Define crossover strategy (DE mutation is included)
        self.crossover = DEX(variant=crossover_variant, CR=CR, at_least_once=True)

        # Define posterior mutation strategy and repair
        self.genetic_mutation = genetic_mutation if genetic_mutation is not None else NoMutation()

    def _do(
        self,
        problem: Problem,
        pop: Population,
        n_offsprings: int,
        **kwargs
    ):

        # Select parents including donor vector
        parents = self.selection(
            problem, pop,
            n_offsprings, self.n_parents,
            to_pop=False, **kwargs,
        )

        # Mutant vectors from DE
        mutants = self.de_mutation(problem, pop, parents, **kwargs)

        # Crossover between mutant vector and target
        matings = self.merge_columnwise(pop, mutants)
        trials = self.crossover(problem, matings, **kwargs)

        # Perform posterior mutation and repair if passed
        off = self.genetic_mutation(problem, trials, **kwargs)

        return off

    @property
    def n_parents(self):
        return self.de_mutation.n_parents

    @staticmethod
    def merge_columnwise(*populations):
        matings = np.column_stack(populations).view(Population)
        return matings



def _fix_deprecated_pm_kwargs(kwargs, genetic_mutation):
    if "pm" in kwargs:
        warnings.warn(
            "pm is deprecated; use 'genetic_mutation'",
            DeprecationWarning, 2
        )
        if genetic_mutation is None:
            genetic_mutation = kwargs["pm"]
    elif "mutation" in kwargs:
        warnings.warn(
            "mutation is deprecated; use 'genetic_mutation'",
            DeprecationWarning, 2
        )
        if genetic_mutation is None:
            genetic_mutation = kwargs["mutation"]

    return kwargs, genetic_mutation



# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True

import numpy as np

from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.set cimport set as cpp_set


cdef extern from "math.h":
    double HUGE_VAL


def calc_mnn(double[:, :] X, int n_remove=0):

    cdef:
        int N, M, n
        cpp_set[int] extremes
        vector[int] extremes_min, extremes_max

    N = X.shape[0]
    M = X.shape[1]

    if n_remove <= (N - M):
        if n_remove < 0:
            n_remove = 0
        else:
            pass
    else:
        n_remove = N - M

    extremes_min = c_get_argmin(X)
    extremes_max = c_get_argmax(X)

    extremes = cpp_set[int]()

    for n in extremes_min:
        extremes.insert(n)

    for n in extremes_max:
        extremes.insert(n)
    
    X = c_normalize_array(X, extremes_max, extremes_min)

    return c_calc_mnn(X, n_remove, N, M, extremes)


def calc_2nn(double[:, :] X, int n_remove=0):

    cdef:
        int N, M, n
        cpp_set[int] extremes
        vector[int] extremes_min, extremes_max

    N = X.shape[0]
    M = X.shape[1]

    if n_remove <= (N - M):
        if n_remove < 0:
            n_remove = 0
        else:
            pass
    else:
        n_remove = N - M

    extremes_min = c_get_argmin(X)
    extremes_max = c_get_argmax(X)

    extremes = cpp_set[int]()

    for n in extremes_min:
        extremes.insert(n)

    for n in extremes_max:
        extremes.insert(n)

    X = c_normalize_array(X, extremes_max, extremes_min)
    
    M = 2

    return c_calc_mnn(X, n_remove, N, M, extremes)


cdef c_calc_mnn(double[:, :] X, int n_remove, int N, int M, cpp_set[int] extremes):

    cdef:
        int n, n_removed, k
        cpp_set[int] calc_items
        cpp_set[int] H
        double[:, :] Dnn, Dij, upper_square
        double[:] d
        int[:, :] Mnn
    
    #Define items to calculate distances
    calc_items = cpp_set[int]()
    for n in range(N):
        calc_items.insert(n)
    for n in extremes:
        calc_items.erase(n)
    
    #Define remaining items to evaluate
    H = cpp_set[int]()
    for n in range(N):
        H.insert(n)
    
    #Upper limits for squared l2 distances
    _Xsum = np.sum(X, axis=1, keepdims=True)
    _upper_square = (_Xsum - _Xsum.T) * (_Xsum - _Xsum.T) / M
    upper_square = _upper_square[:, :]

    #Initialize
    n_removed = 0

    #Initialize neighbors and distances
    _Mnn = np.full((N, M), 0, dtype=np.intc)
    _Dnn = np.full((N, M), HUGE_VAL, dtype=np.double)
    _Dij = np.full((N, N), -1.0, dtype=np.double)
    dd = np.full((N,), HUGE_VAL, dtype=np.double)

    Mnn = _Mnn[:, :]
    Dnn = _Dnn[:, :]
    Dij = _Dij[:, :]
    d = dd[:]

    #Fill in neighbors and distance matrix
    c_calc_mnn_iter(
            X,
            Mnn,
            Dnn,
            Dij,
            N, M,
            calc_items,
            H,
            upper_square
        )

    #Obtain distance metrics
    c_calc_d(d, Dnn, calc_items, M)

    #While n_remove not acheived
    while n_removed < n_remove:

        #Obtain element to drop
        k = c_get_drop(d, H)
        H.erase(k)

        #Update index
        n_removed = n_removed + 1
        if n_removed == n_remove:
            break

        else:

            #Get items to be recalculated
            calc_items = c_get_calc_items(Mnn, Dnn, H, k, M)
            for n in extremes:
                calc_items.erase(n)
            
            #Fill in neighbors and distance matrix
            c_calc_mnn_iter(
                    X,
                    Mnn,
                    Dnn,
                    Dij,
                    N, M,
                    calc_items,
                    H,
                    upper_square
                )

            #Obtain distance metrics
            c_calc_d(d, Dnn, calc_items, M)

    return dd


cdef c_calc_mnn_iter(
    double[:, :] X,
    int[:, :] Mnn,
    double[:, :] Dnn,
    double[:, :] Dij,
    int N, int M,
    cpp_set[int] calc_items,
    cpp_set[int] H,
    double[:, :] upper_square
    ):

    cdef:
        int i, j, m, MM
        double dij
    
    #MM might be different from M in 2NN
    MM = X.shape[1]
    
    #Iterate over items to calculate
    for i in calc_items:

        #Iterate over elements in X
        for j in H:
            
            #Calculate distance if elements satisfy rule
            if ((j != i)
                and (upper_square[i, j] <= Dnn[i, M-1])
                and (Dij[i, j] <= Dnn[i, M-1])):
                
                #Calculate if different from -1
                if Dij[i, j] == -1.0:

                    #Squared distance
                    dij = 0
                    for m in range(MM):
                        dij = dij + (X[j, m] - X[i, m]) * (X[j, m] - X[i, m])
                    
                    #Fill values
                    Dij[i, j] = dij
                    Dij[j, i] = Dij[i, j]

                else:
                    dij = Dij[i, j]
                
                #Iterate over current values
                for m in range(M):
                    
                    #Check is any should be replaced
                    if (dij <= Dnn[i, m]):
                        
                        #Replace higher values
                        Dnn[i, m + 1:] = Dnn[i, m:-1]
                        Mnn[i, m + 1:] = Mnn[i, m:-1]
                        
                        #Replace current value
                        Dnn[i, m] = dij
                        Mnn[i, m] = j
                        
                        break


#Calculate crowding metric
cdef c_calc_d(double[:] d, double[:, :] Dnn, cpp_set[int] calc_items, int M):

    cdef:
        int i, m
    
    for i in calc_items:

        d[i] = 1
        for m in range(M):
            d[i] = d[i] * Dnn[i, m]


#Returns indexes of items to be recalculated after removal
cdef cpp_set[int] c_get_calc_items(
    int[:, :] Mnn,
    double[:, :] Dnn,
    cpp_set[int] H,
    int k, int M):

    cdef:
        int i, m
        cpp_set[int] calc_items
    
    calc_items = cpp_set[int]()

    for i in H:

        for m in range(M):

            if Mnn[i, m] == k:

                Dnn[i, :] = HUGE_VAL
                calc_items.insert(i)
    
    return calc_items


#Returns elements to remove based on crowding metric d and heap of remaining elements H
cdef int c_get_drop(double[:] d, cpp_set[int] H):

    cdef:
        int i, min_i
        double min_d

    min_d = HUGE_VAL
    min_i = 0

    for i in H:

        if d[i] <= min_d:
            min_d = d[i]
            min_i = i
    
    return min_i


#Returns vector of positions of minimum values along axis 0 of a 2d memoryview
cdef vector[int] c_get_argmin(double[:, :] X):

    cdef:
        int N, M, min_i, n, m
        double min_val
        vector[int] indexes
    
    N = X.shape[0]
    M = X.shape[1]

    indexes = vector[int]()
    
    for m in range(M):

        min_i = 0
        min_val = X[0, m]

        for n in range(N):

            if X[n, m] < min_val:

                min_i = n
                min_val = X[n, m]
        
        indexes.push_back(min_i)
    
    return indexes


#Returns vector of positions of maximum values along axis 0 of a 2d memoryview
cdef vector[int] c_get_argmax(double[:, :] X):

    cdef:
        int N, M, max_i, n, m
        double max_val
        vector[int] indexes
    
    N = X.shape[0]
    M = X.shape[1]

    indexes = vector[int]()
    
    for m in range(M):

        max_i = 0
        max_val = X[0, m]

        for n in range(N):

            if X[n, m] > max_val:

                max_i = n
                max_val = X[n, m]
        
        indexes.push_back(max_i)
    
    return indexes


#Performs normalization of a 2d memoryview
cdef double[:, :] c_normalize_array(double[:, :] X, vector[int] extremes_max, vector[int] extremes_min):

    cdef:
        int N = X.shape[0]
        int M = X.shape[1]
        int n, m, l, u
        double l_val, u_val, diff_val
        vector[double] min_vals, max_vals
    
    min_vals = vector[double]()
    max_vals = vector[double]()

    m = 0
    for u in extremes_max:
        u_val = X[u, m]
        max_vals.push_back(u_val)
        m = m + 1
    
    m = 0
    for l in extremes_min:
        l_val = X[l, m]
        min_vals.push_back(l_val)
        m = m + 1
    
    for m in range(M):

        diff_val = max_vals[m] - min_vals[m]
        if diff_val == 0.0:
            diff_val = 1.0

        for n in range(N):

            X[n, m] = (X[n, m] - min_vals[m]) / diff_val
    
    return X